#
# Copyright 2020-2022 Picovoice Inc.
#
# You may not use this file except in compliance with the license. A copy of the license is located in the "LICENSE"
# file accompanying this source.
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.
#


import struct
import sys
import unittest
import wave

import pvporcupine

from picovoice import Picovoice
from test_util import *


class PicovoiceTestData:
    def __init__(self):
        self.picovoiceInstance = None
        self.reset()

    def reset(self):
        self.is_wake_word_detected = False
        self.inference = None

    def wake_word_callback(self):
        self.is_wake_word_detected = True

    def inference_callback(self, inference):
        self.inference = inference


class PicovoiceTestCase(unittest.TestCase):
    @staticmethod
    def _concatenate(language, context, keyword):
        return '%s#%s#%s' % (language, context, keyword)

    @staticmethod
    def __read_file(file_name, sample_rate):
        wav_file = wave.open(file_name, mode="rb")
        channels = wav_file.getnchannels()
        num_frames = wav_file.getnframes()

        if wav_file.getframerate() != sample_rate:
            raise ValueError(
                "Audio file should have a sample rate of %d, got %d" % (sample_rate, wav_file.getframerate()))

        samples = wav_file.readframes(num_frames)
        wav_file.close()

        frames = struct.unpack('h' * num_frames * channels, samples)

        if channels == 2:
            print("Picovoice processes single-channel audio but stereo file is provided. Processing left channel only.")

        return frames[::channels]

    @classmethod
    def setUpClass(cls):
        models = [
            ('en', 'coffee_maker', 'picovoice'),
            ('es', 'iluminación_inteligente', 'manzana'),
            ('de', 'beleuchtung', 'heuschrecke')]

        cls._pvTestDataDictionary = dict()
        for model in models:
            language, context, keyword = model
            pvTestData = PicovoiceTestData()

            _picovoiceInstance = Picovoice(
                access_key=sys.argv[1],
                keyword_path=pv_keyword_paths_by_language(language)[keyword],
                context_path=context_path(context, language),
                porcupine_model_path=pv_porcupine_model_path_by_language(language),
                rhino_model_path=pv_rhino_model_path_by_language(language),
                wake_word_callback=pvTestData.wake_word_callback,
                inference_callback=pvTestData.inference_callback)

            pvTestData.picovoiceInstance = _picovoiceInstance
            cls._pvTestDataDictionary[cls._concatenate(language, context, keyword)] = pvTestData

    @classmethod
    def tearDownClass(cls):
        for pvTestData in cls._pvTestDataDictionary.values():
            pvTestData.picovoiceInstance.delete()

    def run_picovoice(self, language, context, keyword, audio_file_name, intent, slots):
        _pvTestData = self._pvTestDataDictionary[self._concatenate(language, context, keyword)]
        _pvTestData.reset()
        _picovoiceInstance = _pvTestData.picovoiceInstance

        audio = \
            self.__read_file(
                os.path.join(os.path.dirname(__file__), '../../resources/audio_samples', audio_file_name),
                _picovoiceInstance.sample_rate)

        for i in range(len(audio) // _picovoiceInstance.frame_length):
            frame = audio[i * _picovoiceInstance.frame_length:(i + 1) * _picovoiceInstance.frame_length]
            _picovoiceInstance.process(frame)

        self.assertTrue(_pvTestData.is_wake_word_detected)
        self.assertEqual(_pvTestData.inference.intent, intent)
        self.assertEqual(_pvTestData.inference.slots, slots)

    def test(self):
        self.run_picovoice(
            language='en',
            context='coffee_maker',
            keyword='picovoice',
            audio_file_name='picovoice-coffee.wav',
            intent='orderBeverage',
            slots=dict(size='large', beverage='coffee'))

    def test_again(self):
        self.test()

    def test_es(self):
        self.run_picovoice(
            language='es',
            context='iluminación_inteligente',
            keyword='manzana',
            audio_file_name='manzana-luz_es.wav',
            intent='changeColor',
            slots=dict(location='habitación', color='rosado'))

    def test_de(self):
        self.run_picovoice(
            language='de',
            context='beleuchtung',
            keyword='heuschrecke',
            audio_file_name='heuschrecke-beleuchtung_de.wav',
            intent='changeState',
            slots=dict(state='aus'))

    def test_es_again(self):
        self.test_es()

    def test_de_again(self):
        self.test_de()


if __name__ == '__main__':
    if len(sys.argv) != 2:
        print("usage: test_porcupine.py ${ACCESS_KEY}")
        exit(1)

    unittest.main(argv=sys.argv[:1])



#
# Copyright 2020-2022 Picovoice Inc.
#
# You may not use this file except in compliance with the license. A copy of the license is located in the "LICENSE"
# file accompanying this source.
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.
#

import logging
import os
import platform
import subprocess


log = logging.getLogger('PPN')
log.setLevel(logging.WARNING)


def _pv_linux_machine(machine):
    if machine == 'x86_64':
        return machine
    elif machine == 'aarch64':
        arch_info = '-' + machine
    elif machine in ['armv7l', 'armv6l']:
        arch_info = ''
    else:
        raise NotImplementedError("Unsupported CPU architecture: '%s'" % machine)

    cpu_info = ''
    try:
        cpu_info = subprocess.check_output(['cat', '/proc/cpuinfo']).decode()
        cpu_part_list = [x for x in cpu_info.split('\n') if 'CPU part' in x]
        cpu_part = cpu_part_list[0].split(' ')[-1].lower()
    except Exception as error:
        raise RuntimeError("Failed to identify the CPU with '%s'\nCPU info: %s" % (error, cpu_info))

    if '0xb76' == cpu_part:
        return 'arm11' + arch_info
    elif '0xc07' == cpu_part:
        return 'cortex-a7' + arch_info
    elif '0xd03' == cpu_part:
        return 'cortex-a53' + arch_info
    elif '0xd07' == cpu_part:
        return 'cortex-a57' + arch_info
    elif '0xd08' == cpu_part:
        return 'cortex-a72' + arch_info
    elif '0xc08' == cpu_part:
        return 'beaglebone' + arch_info
    elif machine == 'armv7l':
        log.warning(
            'WARNING: Please be advised that this device (CPU part = %s) is not officially supported by Picovoice. '
            'Falling back to the armv6-based (Raspberry Pi Zero) library. This is not tested nor optimal.' % cpu_part)
        return 'arm11'
    else:
        raise NotImplementedError("Unsupported CPU: '%s'." % cpu_part)


def _pv_platform():
    pv_system = platform.system()
    if pv_system not in {'Darwin', 'Linux', 'Windows'}:
        raise ValueError("Unsupported system '%s'." % pv_system)

    if pv_system == 'Linux':
        pv_machine = _pv_linux_machine(platform.machine())
    else:
        pv_machine = platform.machine()

    return pv_system, pv_machine


_PV_SYSTEM, _PV_MACHINE = _pv_platform()

_RASPBERRY_PI_MACHINES = {'arm11', 'cortex-a7', 'cortex-a53', 'cortex-a72', 'cortex-a53-aarch64', 'cortex-a72-aarch64'}
_JETSON_MACHINES = {'cortex-a57-aarch64'}


def pv_model_path(relative):
    return os.path.join(os.path.dirname(__file__), relative, 'lib/common/porcupine_params.pv')


def pv_keyword_files_subdir():
    if _PV_SYSTEM == 'Darwin':
        return 'mac'
    elif _PV_SYSTEM == 'Linux':
        if _PV_MACHINE == 'x86_64':
            return 'linux'
        elif _PV_MACHINE in _JETSON_MACHINES:
            return 'jetson'
        elif _PV_MACHINE in _RASPBERRY_PI_MACHINES:
            return 'raspberry-pi'
        elif _PV_MACHINE == 'beaglebone':
            return 'beaglebone'
    elif _PV_SYSTEM == 'Windows':
        return 'windows'

    raise NotImplementedError('Unsupported platform')


def __append_language(s, language):
    if language == 'en':
        return s
    return '%s_%s' % (s, language)


def context_path(context, language):
    system = platform.system()

    contexts_root = __append_language('../../resources/rhino/resources/contexts', language)

    if system == 'Darwin':
        return os.path.join(os.path.dirname(__file__), contexts_root, 'mac', '%s_mac.rhn' % context)
    elif system == 'Linux':
        if platform.machine() == 'x86_64':
            return os.path.join(os.path.dirname(__file__), contexts_root, 'linux', '%s_linux.rhn' % context)
        else:
            cpu_info = ''
            try:
                cpu_info = subprocess.check_output(['cat', '/proc/cpuinfo']).decode()
                cpu_part_list = [x for x in cpu_info.split('\n') if 'CPU part' in x]
                cpu_part = cpu_part_list[0].split(' ')[-1].lower()
            except Exception as error:
                raise RuntimeError("Failed to identify the CPU with '%s'\nCPU info: %s" % (error, cpu_info))

            if '0xb76' == cpu_part or '0xc07' == cpu_part or '0xd03' == cpu_part or '0xd08' == cpu_part:
                return os.path.join(os.path.dirname(__file__),
                                    contexts_root, 'raspberry-pi', '%s_raspberry-pi.rhn' % context)
            elif '0xd07' == cpu_part:
                return os.path.join(os.path.dirname(__file__),
                                    contexts_root, 'jetson', '%s_jetson.rhn' % context)
            elif '0xc08' == cpu_part:
                return os.path.join(os.path.dirname(__file__),
                                    contexts_root, 'beaglebone', '%s_beaglebone.rhn' % context)
            else:
                raise NotImplementedError("Unsupported CPU: '%s'." % cpu_part)
    elif system == 'Windows':
        return os.path.join(os.path.dirname(__file__), contexts_root, 'windows', '%s_windows.rhn' % context)
    else:
        raise ValueError("Unsupported system '%s'." % system)


def pv_keyword_paths_by_language(language):
    keyword_files_root = __append_language('resources/keyword_files', language)
    relative = '../../resources/porcupine'
    keyword_files_dir = \
        os.path.join(os.path.dirname(__file__), relative, keyword_files_root, pv_keyword_files_subdir())

    res = dict()
    for x in os.listdir(keyword_files_dir):
        res[x.rsplit('_')[0]] = os.path.join(keyword_files_dir, x)

    return res


def pv_rhino_model_path_by_language(language):
    model_path_subdir = __append_language('lib/common/rhino_params', language)
    model_path_subdir = '%s.pv' % model_path_subdir
    return os.path.join(os.path.dirname(__file__), '../../resources/rhino', model_path_subdir)


def pv_porcupine_model_path_by_language(language):
    model_path_subdir = __append_language('lib/common/porcupine_params', language)
    model_path_subdir = '%s.pv' % model_path_subdir
    return os.path.join(os.path.dirname(__file__), '../../resources/porcupine', model_path_subdir)
